/**
* Copyright (c) 2010, Leonardo R. Bachega and Charles A. Bouman
* All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the Purdue University nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Leonardo R. Bachega and Charles A. Bouman BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
import java.io.*;
import java.util.*;

public class FastSMT4 extends FastSMT
{
	int[] MaxCostNeighbor;
	/* main data structures*/
	int p; //number of dimensions 
	int n; //number of samples
	int M=8; //max. number of neighbors for each point
	double[][] Y; //data vector
	//int[] MaxCostNeighbor;
	
	// Required minimum value for eigenvalue estimates
	// See Guangzhi et al. on IEEE Transactions on Image Processing, March 2011 for further discussion
	double eigenvalue_floor = 0; //eigenvalue floor default value is zero

	HashSet<Integer> NeighborSet[];	//set each point's neighbors
	HashSet<Integer> PointedBySet[]; //set of all elements that points to i
	HashMap<Integer, Double> MyNeighborCosts[]; //keep track of the cost assoc. to the correlation to each of the neighbors
	ArrayList<Rotation> RotationList = null; //list of rotations
	TreeSet<Integer> ElementSet; //all elements in a data structure searchable in O(log p)
	double[] MaxCost;
	//int[] MaxCostNeighbor;

	boolean pruneNeighborList = false; //flag to prune neighborhoods

	/*data structures to support performance characterization / debug / consistency check*/
	double currentFrobeniusNorm, currentTrace;
	boolean FrobeniusNormComputed = false;
	boolean TraceComputed = false;
	int[] numTouches = null; //number of times a point is touched
	MethodCounter mc;
	
	/**
	 * Initialize all the data structures
	 * pre-condition: value of p is known
	 **/
	private void init(int p)
	{
		this.p = p;
		NeighborSet = new HashSet[this.p];
		numTouches = new int[this.p]; 
		PointedBySet = new HashSet[this.p];
		MyNeighborCosts = new HashMap[this.p];
		
		MaxCost = new double[this.p];
		
		
		for (int i=0; i < this.p ; i++)
		{
			this.NeighborSet[i] = new HashSet<Integer>();
			this.PointedBySet[i] = new HashSet<Integer>();
			this.MyNeighborCosts[i] = new HashMap<Integer, Double>();
			this.numTouches[i] = 0;
		}
		
		MaxCostNeighbor = new int[this.p];
		for (int i=0; i < this.p; i++)
		{
			this.MaxCostNeighbor[i] = 0;
		}
		
		this.RotationList = new ArrayList<Rotation>();
		
		/*For performance debugging*/
		this.mc = new MethodCounter();
		this.mc.count ("init");
		return;
	}
	
	/**
	 * Construct the SMT without initializing the data vector Y 
	 * p is the number of dimensions
	 * This should be used only to test the neighborhood operations and not in any 
	 * meaningful SMT operation
	 **/
	public FastSMT4(int p)
	{
		this.init(p);
	}
	
	public FastSMT4(double[][] Y)
	{
		this.Y = Y;
		int p = Y.length;
		this.n = Y[0].length;
		init (p);
	}

	/**
	 * Compute the inner product of two vectors in double format and returns the result
	 */
	public double ComputeDotProduct(double x[], double y[])
	{
		double sum=0.0;
		
		for (int i=0; i< x.length;i++)
		{
			sum += x[i]*y[i];
		}
		return sum;
	}
	public void pruneNeighborhoodList(int M)
	{
		this.pruneNeighborList = true;
		this.M = M;
		return;
	}
	public void pruneNeighborhoodList()
	{	
		this.pruneNeighborList = true;
		return;
	}
	/*
	 *  Set floor for the values of estimated eigenvalues
	 */
	public void addEigenvalueFloor (double eigenvalue_floor)
	{
		this.eigenvalue_floor = eigenvalue_floor;
		return;
	}
	public double getEigenvalueFloor ()
	{
		return this.eigenvalue_floor;
	}

	/**
	 * Initialize the neighborhood sets based on the pixel neighborhood size=1,2,3 ...
	 * size=1 is equivalent of selecting as neighbors the 8 closest pixels, etc...
	 * Assumes the image vectors are stored in row-major format (raster column order) and the
	 * image height and width are both known
	 */
	public void addImageNeighbors(int height, int width, int size)
	{
		
		for (int pixel=0; pixel< this.p;pixel++)
		{
			int ipixel, jpixel;
			int istart, jstart, iend, jend;
			
			ipixel = pixel / height;
			jpixel = pixel % height;
			
			istart = Math.max(ipixel - size, 0);
			jstart = Math.max(jpixel - size, 0);
			
			iend = Math.min(ipixel + size, width-1);
			jend = Math.min(jpixel + size, height-1);
			
			ArrayList<Integer> NeighborList = new ArrayList<Integer>();
			
			for (int i = istart; i <= iend; i++)
			{
				for (int j = jstart; j <= jend; j++)
				{
					this.setAsNeighbors(pixel, i*height+j);
				}
			}
		}
		initialize_max_costs();
	}

	public void setNeighborsMap(int[][] vNeighbors)
	{
		for(int nCoordinateIndex=0; nCoordinateIndex < this.p; nCoordinateIndex++)
		{
			int nNumNeighbors = vNeighbors[nCoordinateIndex].length;
			for (int nNeighborIndex=0; nNeighborIndex < nNumNeighbors; nNeighborIndex ++)
			{
				if (vNeighbors[nCoordinateIndex][nNeighborIndex] != -1)
					this.setAsNeighbors(nCoordinateIndex, vNeighbors[nCoordinateIndex][nNeighborIndex]-1);
			}
		}
		initialize_max_costs();
		return;
	}
	
	/**
	 * Initialize the neighborhood sets by setting each point as the neighbor of all other
	 * points
	 */
	public void addAllNeighbors()
	{
		//System.err.println("p="+this.p);
		for (int i=0; i< this.p; i++)
			for (int j=0; j< i; j++)
			{
				//System.err.println("Set i="+i+" ,j="+j);
				this.setAsNeighbors(i, j);
			}
		initialize_max_costs();
	}

	//later generalize for an arbitrary length
	public void addCloseIndexNeighbors()
	{
		this.setAsNeighbors(0, 1);
		
		for (int i=1;i <this.p-1; i++)
		{
			this.setAsNeighbors(i, i+1);
			this.setAsNeighbors(i, i-1);
		}
		
		this.setAsNeighbors(this.p-1, this.p-2);
		initialize_max_costs();
	}
	
	/**
	 * Return true if the points are neighbors of each other and false if not
	 * this assumes symmetric neighborhoods: i is neighbor of j <=> j is neighbor of i
	 */
	boolean CheckForNeighbors(int i, int j)
	{
		return  (this.NeighborSet[j].contains(i) && this.NeighborSet[i].contains(j));
	}
	
	void add_all_related_elements_to_smart_set(int i, int j)
	{
		this.mc.count ("rb_tree.add");
		this.mc.count ("rb_tree.add");
		this.ElementSet.add(i);
		this.ElementSet.add(j);

		Iterator<Integer> itI = this.PointedBySet[i].iterator();
		while (itI.hasNext())
		{
			int l = itI.next();
			this.ElementSet.add(l);
			this.mc.count ("rb_tree.add");
		}

		Iterator<Integer> itJ = this.PointedBySet[j].iterator();
		while(itJ.hasNext())
		{
			int l = itJ.next();
			this.ElementSet.add(l);
			this.mc.count ("rb_tree.add");
		}
		return;
	}


	void remove_all_related_elements_from_smart_set(int i, int j)
	{
		this.mc.count ("rb_tree.remove");
		this.mc.count ("rb_tree.remove");
		this.ElementSet.remove(i);
		this.ElementSet.remove(j);

		Iterator<Integer> itI = this.PointedBySet[i].iterator();
		while (itI.hasNext())
		{
			int l = itI.next();
			this.ElementSet.remove(l);
			this.mc.count ("rb_tree.remove");
		}

		Iterator<Integer> itJ = this.PointedBySet[j].iterator();
		while(itJ.hasNext())
		{
			int l = itJ.next();
			this.ElementSet.remove(l);
			this.mc.count ("rb_tree.remove");
		}
		return;
	}

	void mergeNeighborSets2(int i, int j)
	{
		// define order relation for the neighbors of a point as the difference
		// in cost 
		CostComparator Icomp = new CostComparator(i, this);
		CostComparator Jcomp = new CostComparator(j, this);
		
		HashSet<Integer> newIneighbors = new HashSet<Integer>();
		HashSet<Integer> newJneighbors = new HashSet<Integer>();
		
		HashSet<Integer> NeighborsCombined = new HashSet<Integer>();
		
		List<Integer> newINeighborCandidateList = new ArrayList<Integer>();
		List<Integer> newJNeighborCandidateList = new ArrayList<Integer>();
		
		HashMap<Integer, Double> newINeighborCosts = new HashMap<Integer, Double>();
		HashMap<Integer, Double> newJNeighborCosts = new HashMap<Integer, Double>();

		remove_all_related_elements_from_smart_set(i,j);
		
		NeighborsCombined.addAll(this.NeighborSet[i]);
		NeighborsCombined.addAll(this.NeighborSet[j]);
		
		//create list for i
		NeighborsCombined.remove(i);
		newINeighborCandidateList.addAll(NeighborsCombined);
		Collections.sort(newINeighborCandidateList, Icomp);
		
		for (int k=0; k < newINeighborCandidateList.size(); k++)
		{
			Integer candidate = newINeighborCandidateList.get(k);
			if (k < this.M)
			{
				//add neighbor to i's list
				newIneighbors.add(candidate);
				this.PointedBySet[candidate].add(i);
				newINeighborCosts.put(candidate, Cost(candidate, i));
			}
			else
			{
				this.PointedBySet[candidate].remove(i);
			}
		}
		
		//create list for j
		NeighborsCombined.add(i);
		NeighborsCombined.remove(j);
		newJNeighborCandidateList.addAll(NeighborsCombined);
		Collections.sort(newJNeighborCandidateList, Jcomp);
		
		for (int k=0; k < newJNeighborCandidateList.size(); k++)
		{
			Integer candidate = newJNeighborCandidateList.get(k);
			if (k < this.M)
			{
				//add neighbor to i's list
				newJneighbors.add(candidate);
				this.PointedBySet[candidate].add(j);
				newJNeighborCosts.put(candidate, Cost(candidate, j));
			}
			else
			{
				this.PointedBySet[candidate].remove(j);
			}
		}
		
		this.NeighborSet[i] = newIneighbors;
		this.NeighborSet[j] = newJneighbors;

		this.MyNeighborCosts[i] = newINeighborCosts;
		this.MyNeighborCosts[j] = newJNeighborCosts;
		
		/*update the max cost and max neighbor for both i and j*/
		
		int aux = selectMaxCostNeighbor(i);
		this.MaxCostNeighbor[i] = aux;
		this.MaxCost[i] = this.MyNeighborCosts[i].get(this.MaxCostNeighbor[i]);
		
		aux = selectMaxCostNeighbor(j);
		this.MaxCostNeighbor[j] = aux;
		this.MaxCost[j] = this.MyNeighborCosts[j].get(this.MaxCostNeighbor[j]);

		/*look at all elements that point to i and j and recompute their costs*/
		
		Iterator<Integer> itI = this.PointedBySet[i].iterator();
		while (itI.hasNext())
		{
			int l = itI.next();
			double cost = Cost (i,l);
			this.MyNeighborCosts[l].put(i, cost);
			update_max_cost (l, i, cost);
		}

		Iterator<Integer> itJ = this.PointedBySet[j].iterator();
		while(itJ.hasNext())
		{
			int l = itJ.next();
			double cost = Cost (j,l);
			this.MyNeighborCosts[l].put(j, cost);
			update_max_cost (l, j, cost);
		}
		
		add_all_related_elements_to_smart_set(i,j);
	}
	
	void initialize_max_costs()
	{
		for (int i=0; i< this.p; i++)
		{
			this.MaxCostNeighbor[i] = 1;
		}
		for (int i=0; i< this.p; i++)
		{
			int aux = selectMaxCostNeighbor(i);
			this.MaxCostNeighbor[i] = aux;
			this.MaxCost[i] = this.MyNeighborCosts[i].get(this.MaxCostNeighbor[i]);
		}
		
		//After all the costs are computed, add to "smart dataset"
		create_smart_neighbor_set();
	}

	void create_smart_neighbor_set()
	{
		ElementCost element_cost = new ElementCost(this);
		this.ElementSet = new TreeSet<Integer>(element_cost);
		for (int i=0; i< this.p; i++)
		{
			this.ElementSet.add(i);
			this.mc.count ("rb_tree.add");
		}
	}

	void update_max_cost(int element, int neighbor, double cost)
	{
		int current_max_neighbor = MaxCostNeighbor[element];
		double current_max_cost = MaxCost[element];
		if ((current_max_neighbor == neighbor && cost < current_max_cost))
		{
			int new_max_neighbor = selectMaxCostNeighbor(element);
			double new_max_cost = this.MyNeighborCosts[element].get(new_max_neighbor);
			this.MaxCostNeighbor[element] = new_max_neighbor;
			this.MaxCost[element] = new_max_cost;
			//search for new max among neighbors
		}
		else if (cost > current_max_cost)
		{
			//we have a new max
			this.MaxCostNeighbor[element] = neighbor;
			this.MaxCost[element] = cost;
		}
		return;
	}
	/**
	 * Merge the NeighborSets of two points (doesn't assume the 2 points must be neighbors, and 
	 * don't make them neighbors if they are not
	 */
	void mergeNeighborSets(int i, int j)
	{
		
		HashSet<Integer> NewINeighbors = new HashSet<Integer>();
		HashSet<Integer> NewJNeighbors = new HashSet<Integer>();
		
		Iterator<Integer> itIneighbors = this.NeighborSet[i].iterator();
		while (itIneighbors.hasNext())
		{
			Integer element = itIneighbors.next();
			if (!NeighborSet[j].contains(element) && element != j)
			{
				NewJNeighbors.add(new Integer(element));
			}
		}
		
		Iterator<Integer> itJneighbors = this.NeighborSet[j].iterator();
		while (itJneighbors.hasNext())
		{
			Integer element = itJneighbors.next();
			if (!NeighborSet[i].contains(element) && element != i)
			{
				NewINeighbors.add(new Integer(element));
			}
		}
		
		Iterator<Integer> itI = NewINeighbors.iterator();
		while (itI.hasNext())
		{
			Integer element = itI.next();
			this.NeighborSet[element].add(new Integer(i));
		}
		
		Iterator<Integer> itJ = NewJNeighbors.iterator();
		while (itJ.hasNext())
		{
			Integer element = itJ.next();
			this.NeighborSet[element].add(new Integer(j));
		}
		
		this.NeighborSet[i].addAll(NewINeighbors);
		this.NeighborSet[j].addAll(NewJNeighbors);
		
		return;
	}
	
	/**
	 * Compute angle theta based on the covariances of the points: Sii, Sjj, Sij
	 */
	double ComputeTheta (double Sii, double Sjj, double Sij)
	{
		return 0.5*Math.atan2(-2*Sij,(Sii-Sjj));
	}
	
	/**
	 * search for the points with highest correlation among all the neighbors
	 */
	int[] selecMaxCostPair()
	{
		double maxCost = 0;
		int[] maxCostPair = {-1,-1}; //new int[2];
		
		this.mc.count ("selectMaxCostPair");
		
		for (int i=0; i< p; i++)
		{
			int l = selectMaxCostNeighbor2(i);
			//System.out.println (i+" Max cost neighbor:"+l);
			double ilCost = getMaxCost(i); //Cost(i,l);
			if (ilCost > maxCost)
			{
				maxCost = ilCost;
				maxCostPair[0] = i;
				maxCostPair[1] = l;
			}
		}
		if (maxCostPair[0] > maxCostPair[1])
		{
			int aux = maxCostPair[0];
			maxCostPair[0] = maxCostPair[1];
			maxCostPair[1] = aux;
		}
		return maxCostPair;
	}
	
	int[] selecMaxCostPair2()
	{
		this.mc.count ("selectMaxCostPair2");
		int[] maxCostPair = { -1, -1};
		int i = ElementSet.last();
		this.mc.count ("rb_tree.getmax");
		int j = MaxCostNeighbor[i];

		if (i < j)
		{
			maxCostPair[0] = i;
			maxCostPair[1] = j;
		}
		else
		{
			maxCostPair[0] = j;
			maxCostPair[1] = i;
		}
		return maxCostPair;
	}
	
	int selectMaxCostNeighbor2(int i)
	{
		return this.MaxCostNeighbor[i];
	}
	
	double getMaxCost(int i)
	{
		return this.MaxCost[i];
	}

	int selectMaxCostNeighbor(int i)
	{
		double maxCost = -1;
		int neighbor = -1;

		Iterator<Integer> it = this.MyNeighborCosts[i].keySet().iterator();
		while (it.hasNext())
		{
			int l = it.next();
			double cost = this.MyNeighborCosts[i].get(l);
			
			if (cost > maxCost)
			{
				neighbor = l;
				maxCost = cost;
			}
		}
		if (neighbor == -1)
		{
			System.err.println ("INSIDE SELECT_MAX_COST_NEIGHBOR: returning the wrong value");
			System.err.println("MaxCost:"+maxCost);
		}
		return neighbor;
	}
	
	/**
	 * Compute the cost(correlation coefficient) between points i and j 
	 */
	double Cost(int i, int j)
	{
		double correlation;
		double Sii, Sij, Sjj;
		
		Sii = S(i,i);
		Sjj = S(j,j);
		Sij = S(i,j);
		
		correlation = (Sij*Sij) / ((Sii+(this.eigenvalue_floor))*(Sjj+(this.eigenvalue_floor)));
		//System.out.println(correlation);
		return correlation;
	}

	/**
	 * Compute the covariance between i and j
	 * The covariance is computed on-the-fly based on the values of Y(i,1:n) and Y(j,1:n)
	 */
	double S(int i, int j)
	{
		double Sij;
		Sij = ComputeDotProduct(Y[i], Y[j])/n;
		return Sij;
	}
	
	public void MatlabApplyRotations(double[][] Rotations)
	{
		int NumRotations = Rotations.length;
		for (int k=0; k < NumRotations; k++)
		{
			
			double dik = Rotations[k][0] - 1;
			double djk = Rotations[k][1] - 1;
			int ik = (int)dik;
			int jk = (int)djk;
			double thetak = Rotations[k][2];
			applyRotation(ik, jk, thetak);
		}
		return;
	}	
	/**
	 * Apply one Given's rotation to the pair (i,j) with angle theta:
     * make Y = Ek'Y
	 *	Y(i,k) = Y(i,k)cost - Y(j,k)sint
	 *	Y(j,k) = Y(i,k)sint + Y(j,k)cost
	 */
	void applyRotation(int i, int j, double theta)
	{
		int L = Y[i].length;
		double Yik,Yjk;
		
		double sint = Math.sin(theta);
		double cost = Math.cos(theta);

		for (int k=0; k < L; k++)
		{
			Yik = Y[i][k];
			Yjk = Y[j][k];
			Y[i][k] = Yik*cost - Yjk*sint;
			Y[j][k] = Yik*sint + Yjk*cost;
		}
		
		this.numTouches[i]++;
		this.numTouches[j]++;
		
		return;
	}
	
	public double[] getRank()
	{
		double [] rank = new double[this.p];
		for (int i=0; i< this.p; i++)
		{
			rank[i] = 0;
			for (int j=0; j < this.p; j++)
			{
				if (this.NeighborSet[j].contains(i))
				{
					rank[i]++;
				}
			}
		}
		return rank;
	}

	public double[] getDiagonal()
	{
		double[] diag = new double[this.p];
		for (int i=0; i< this.p; i++)
		{
			diag[i] = this.S(i, i);
		}
		return diag;
	}
	
	void applyRotation (Rotation rotation)
	{
		this.applyRotation(rotation.getIk(), rotation.getJk(), rotation.getTheta());
	}
	
	/**
	 * Perform a sequence of n Given's rotations
	 */
	public void PerformRotations(int n)
	{
		for (int k=0;k < n; k++)
			PerformRotation();
	}
	
	
	/**
	 * Perform one Given's rotation by selecting the most correlated neighbor points (i,j), computing the angle
	 * and applying the rotation. The neighborhoods of both points are merged 
	 */
	void PerformRotation()
	{
		int[] pair = this.selecMaxCostPair2();
		
		int i = pair[0];
		int j = pair[1];
		
		double theta = this.ComputeTheta(S(i,i), S(j,j), S(i,j));

		this.applyRotation(i, j, theta);
		
		if (this.pruneNeighborList)
			this.mergeNeighborSets2(i, j);
		else
			this.mergeNeighborSets(i, j);
		

		//update rotation list
		RotationList.add(new Rotation(i,j, theta));
		
		return;
	}
	
	/**
	 * Set points i and j as neighbors by adding i to neighborSet(j) and j to neighborSet(i)
	 */
	void setAsNeighbors(int i, int j)
	{
		double cost = Cost(i,j);
		if (i != j)
		{
			this.NeighborSet[i].add(new Integer(j));
			this.MyNeighborCosts[i].put(j, cost);
			this.PointedBySet[i].add(j);

			this.NeighborSet[j].add(new Integer(i));
			this.MyNeighborCosts[j].put(i, cost);
			this.PointedBySet[j].add(i);
		}
		return;
	}
	
	void removeNeighbors(int i, int j)
	{
		if (i!=j)
		{
			this.NeighborSet[i].remove(j);
			this.NeighborSet[j].remove(i);
		}
	}
	/* Print methods*/
	
	/**
	 * Print all the sets of neighbors
	 */
	void printNeighbors()
	{
		for (int i=0;i<p;i++)
		{
			System.out.println(i+" neighbors :"+this.NeighborSet[i].toString());
		}
	}
	
	void printNumNeighbors()
	{
		System.out.println(this.NeighborSet[0].size());
		for (int i=1;i<p;i++)
		{
			System.out.print(" "+this.NeighborSet[i].size());
		}
		System.out.print("\n");
	}
	
	void printVariances()
	{
		double[] variances = this.getDiagonal();
		System.out.print(variances[0]);
		for (int i=1;i<p;i++)
		{
			System.out.print(" "+variances[i]);
		}
		System.out.print("\n");
	}
	
	public double[][] MatlabGetRotations()
	{
		this.mc.printValues();
		return MatlabGetLastRotations(-1);
	}
	
	public double[][] MatlabGetLastRotations(int n)
	{
		int FirstRotationToGet, LastRotationToGet;
		List<Rotation> RotationList = getRotationList();
		if (n > RotationList.size() || n == -1)
		{
			FirstRotationToGet = 0;
			LastRotationToGet = RotationList.size() - 1;
		}
		else
		{
			FirstRotationToGet = RotationList.size() - n;
			LastRotationToGet = RotationList.size() - 1;
		}
		int NumRotations = LastRotationToGet - FirstRotationToGet + 1;

		double[][] RotationsArray = new double[NumRotations][3];
		for (int index=0; index < NumRotations; index++)
		{
			Rotation rotation = RotationList.get(index+FirstRotationToGet);
			RotationsArray[index][0] = rotation.getIk() + 1;
			RotationsArray[index][1] = rotation.getJk() + 1;

			RotationsArray[index][2] = rotation.getTheta();
		}
		return RotationsArray;
	}
	List<Rotation> getRotationList()
	{
		return this.RotationList;
	}
	
	void printRotations()
	{
		Iterator<Rotation> it = this.RotationList.iterator();
		while (it.hasNext())
			it.next().print();
	}
	String printRotationsToString()
	{
		String output = new String();
		Iterator<Rotation> it = this.RotationList.iterator();
		while (it.hasNext())
		{
			output += it.next().printToString()+"\n";
		}
		return output;
	}
	void printRotationsToFile (String filename)
	{
		try
		{
			BufferedWriter out = new BufferedWriter(new FileWriter(filename));
	        	out.write(this.printRotationsToString());
	        	out.close();
		}
		catch (IOException e)
		{
			System.err.println ("Error while dealing with file "+filename);
		}
	}
	
	void printNumTouches()
	{
		for (int i=0;i<p;i++)
		{
			System.out.print(this.numTouches[i]+" ");
		}
		System.out.print("\n");
	}
	
	/* 
	 Check and verification methods 
	 */
	public int[] getPointedByNumbers()
	{
		int[] PointedBySetSize = new int[this.p];
		for (int i=0; i< this.p; i++)
		{
			PointedBySetSize[i] = PointedBySet[i].size();
		}
		return PointedBySetSize;
	}
	/**
	 * Compare the covariances with a reference covariance matrix stored in txtfile in ascii format
	 */
	
	double ComputeSFrobeniusNorm()
	{
		double FrobeniusNorm2 = 0, FrobeniusNorm;
		for (int i=0; i< this.p; i++)
		{
			for (int j=0; j< this.p; j++)
			{
				double sij = S(i,j);
				FrobeniusNorm2 += sij*sij;
			}
		}
		FrobeniusNorm = Math.sqrt(FrobeniusNorm2);
		if (this.FrobeniusNormComputed == false)
		{
			this.currentFrobeniusNorm = FrobeniusNorm;
			this.FrobeniusNormComputed = true;
		}
		else
		{
			if (Math.abs(this.currentFrobeniusNorm - FrobeniusNorm) > 0.0001)
			{
				System.err.println("Warning: Frobenius Norm differs from previously computed! Should be invariant!");
				System.err.println("Previously Computed Value: "+this.currentFrobeniusNorm);
				System.err.println("Current Value:"+FrobeniusNorm);
			}
			else
			{
				System.out.println("Frobenius Norm remains invariant");
			}
			this.currentFrobeniusNorm = FrobeniusNorm;
		}
		return FrobeniusNorm;
	}
	
	double ComputeSTrace()
	{
		double trace=0;
		for (int i=0; i< this.p; i++)
		{
			trace += S(i,i);
		}
		if (this.TraceComputed == false)
		{
			this.currentTrace = trace;
			this.TraceComputed = true;
		}
		else
		{
			if (Math.abs(this.currentTrace - trace) > 0.0001)
			{
				System.err.println("Warning: Trace of S differs from previously computed! Should be invariant!");
				System.err.println("Previously Computed Value: "+this.currentTrace);
				System.err.println("Current Value:"+this.currentTrace);
			}
			else
			{
				System.out.println("Trace of S remains invariant");	
			}
			this.currentTrace = trace;
		}
		return trace;
	}
	
	double ComputeDiagProduct()
	{
		double prod = Math.log(S(0,0));
		for (int i=1; i< this.p;i++)
			prod += Math.log(S(i,i));
		return prod;
	}	
}
