%%%
% Copyright (c) 2010, Leonardo R. Bachega and Charles A. Bouman
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%     * Redistributions of source code must retain the above copyright
%       notice, this list of conditions and the following disclaimer.
%     * Redistributions in binary form must reproduce the above copyright
%       notice, this list of conditions and the following disclaimer in the
%       documentation and/or other materials provided with the distribution.
%     * Neither the name of the Purdue University nor the
%       names of its contributors may be used to endorse or promote products
%       derived from this software without specific prior written permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
% ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
% WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
% DISCLAIMED. IN NO EVENT SHALL Leonardo R. Bachega and Charles A. Bouman BE LIABLE FOR ANY
% DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
% (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
% LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
% ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
% SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
%%%
function [R, best_alpha, avg_alpha_likelihood] = Build_Shrinkage_model_crossvalid(Y, kfold,  alpha_range, type, r, smt_crossvalid_flag, smt_crossvalid_kfold, M, image_height, image_width)
% Builds a shrinkage model using cross-validation to determine alpha in the
% formula R= (1-alpha)*S + alpha*Target
%
% [R, best_alpha, alpha_likelihoods] = Build_Shrinkage_model_crossvalid(Y, kfold,  alpha_range, type, r, use_crossvalid_flag, smt_crossvalid_kfold)
% 
% Y: p x n matrix containing n p-dimensional samples
%
% kfold: split the  samples in kfold ways
%
% alpha_range: range of values of alpha to be searched (e.g. 0.1:0.1:1)
%
% type: implemented types: 'diagonal','FullSMT','FastSMT' (default is'diagonal')
%
% r: r value if the target is 'FullSMT' or 'FastSMT' (r=K/p, K=number of Givens rotations)
%
% smt_crossvalid_flag: if ='' finds the SMT with K=r*p rotations
%                      if ='crossvalidation' finds the SMT through crossvalidation
%                      searching K in the 0..r*P range
%
% smt_crossvalid_kfold: number of kfold ways to split the dataset to
%                       estimate the SMT model (default value = 2)
%
% M: 
%
% image_height
%
% image_width

    if nargin < 7
        smt_crossvalid_kfold = 2;
    end
    if nargin < 6
        smt_crossvalid_flag = '';
    end
    if nargin < 5
        r = 1.5;
    end
    if nargin < 4
        type = 'diagonal';
    end

    [p,n] = size(Y);
    
    delta = floor (n/kfold);
    n_used = kfold*delta; %consider only an integer multiple of kfold - the leftovers are ignored
    
    alpha_likelihoods = zeros (kfold,length(alpha_range));

    for k=1:kfold
        %disp(['k=',mat2str(k)]);
        fullrange = circshift ((1:n_used)', k*delta);
        test_range = fullrange(1:delta)';
        train_range = fullrange(delta+1:n_used)';
        
        n_train = length(train_range);
        n_test = length (test_range);
        
        Y_test = Y(:, test_range);
        Y_train = Y(:, train_range);
        
        S_train = (1/n_train)*(Y_train*Y_train');
        S_test = (1/n_test)*(Y_test*Y_test');
        
        if (strcmp(type,'diagonal'))
            shrinkage_target = diag(diag (S_train)); %add other targets here
        elseif (strcmp(type,'FullSMT'))
            shrinkage_target = SMT_full_covariance_estimation(Y_train, r, smt_crossvalid_flag, smt_crossvalid_kfold);
        elseif (strcmp(type,'FastSMT'))
            shrinkage_target = SMT_neighborhood_covariance_estimation(Y_train, r, smt_crossvalid_flag, smt_crossvalid_kfold, M, image_height, image_width);
        end
        % find likelihood for each alpha
        for alpha_iter = 1:length(alpha_range)
            alpha = alpha_range (alpha_iter);
            R = alpha * shrinkage_target + (1-alpha) * S_train;
            [E,Lambda]=eig(R);
            alpha_likelihoods (k, alpha_iter) = SMT_compute_likelihood(diag (E'*S_test*E), diag (Lambda), n_test) / n_test;
        end
    end
       
    avg_alpha_likelihood = mean (alpha_likelihoods);
    
    [max_l, max_index] = max (avg_alpha_likelihood);
    best_alpha = alpha_range(max_index);
    
    %Compute the best model using all the available samples
    S = (1/n)*(Y*Y');
    
    if (strcmp(type,'diagonal'))
        shrinkage_target = diag(diag (S)); %add other targets here
    elseif (strcmp(type,'FullSMT'))
        shrinkage_target = SMT_full_covariance_estimation(Y, r, smt_crossvalid_flag, smt_crossvalid_kfold);
    elseif (strcmp(type,'FastSMT'))
        shrinkage_target = SMT_neighborhood_covariance_estimation(Y, r, smt_crossvalid_flag, smt_crossvalid_kfold, M, image_height, image_width);
    end
    
    
    R = best_alpha*shrinkage_target + (1-best_alpha)*S;
    
end