%%%
% Copyright (c) 2010, Leonardo R. Bachega and Charles A. Bouman
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%     * Redistributions of source code must retain the above copyright
%       notice, this list of conditions and the following disclaimer.
%     * Redistributions in binary form must reproduce the above copyright
%       notice, this list of conditions and the following disclaimer in the
%       documentation and/or other materials provided with the distribution.
%     * Neither the name of the Purdue University nor the
%       names of its contributors may be used to endorse or promote products
%       derived from this software without specific prior written permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
% ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
% WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
% DISCLAIMED. IN NO EVENT SHALL Leonardo R. Bachega and Charles A. Bouman BE LIABLE FOR ANY
% DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
% (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
% LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
% ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
% SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
%%%
function   [T,lambda,SMTArray,likelihood, complexity_counter] = SMT_build_model(R,NumRotations,Y)
    % R - Input convariance matrix
    % NumRotations - number of rotations
    % Y - the sample data
    %
    % SMTArray contains the sequence of SMT rotions. So that 
    %           SMTArray(i,1) = first index
    %           SMTArray(i,2) = second index
    %           SMTArray(i,3) = rotation angle
    %
    % T is the product of the SMT rotations, so that
    %           D = T*R*T^t; 
    % where D is approximately diagonal
    % and       lambda = diag(D)
    % likelihood: likelihood of the test data

    complexity_counter = zeros (1,3);

    if nargin == 3
        flag_design = 1; % design process
        M=size(Y,2);
        S=1/M*Y*Y';
    else
        flag_design =0;  % generate the desired SMT covariance estimate
    end
    likelihood =zeros(NumRotations,1);

    T = eye(size(R));
    N = max(size(R));


    SMTArray = zeros(NumRotations,3);

    F = Criteria(R);
    [f,Irow]=max(F);
    for iterations=1:NumRotations

    %%%% slow search %%%%%%%
    %  [f,Irow]=max(F);
    %  [val,Icol]=max(f);
    %  i=Irow(Icol);
    %  j=Icol;

    %%%% fast search %%%%%%%
      [val,Icol]=max(f);
      i=Irow(Icol);
      j=Icol;


      if(i>j) 
        ii = i; i=j; j=ii;
      end

      %disp(sprintf('i=%d,j=%d \n',i,j));

      rii = R(i,i);
      rij = R(i,j);
      rjj = R(j,j);

      [a,b,theta] = rotation(rii,rjj,rij);

      %%% fast implementation %%%%
      % R=A'*R*A 
      ci = R(:,i)*cos(theta)+R(:,j)*sin(theta);   % R*A
      cj = -R(:,i)*sin(theta)+R(:,j)*cos(theta);
      R(:,i) = ci;
      R(:,j) = cj;

      ri = R(i,:)*cos(theta) + R(j,:)*sin(theta); % A'*R*A
      rj = -R(i,:)*sin(theta) + R(j,:)*cos(theta); 
      R(i,:) = ri;  
      R(j,:) = rj;


      % T=T*A
      ri = T(:,i)*cos(theta)+T(:,j)*sin(theta);   % T*A
      rj = -T(:,i)*sin(theta)+T(:,j)*cos(theta);
      T(:,i) = ri;
      T(:,j) = rj;



      %%%%%% update the F matrix %%%%%%
      lambda=diag(R);
      F(:,i)=R(:,i).^2 ./ (R(i,i)*lambda);
      F(:,j)=R(:,j).^2 ./ (R(j,j)*lambda);
      F(i,:)=F(:,i)';
      F(j,:)=F(:,j)';
      F(i,i)=0;
      F(j,j)=0;


      %%%%%% update the f and Irow vectors %%%%%%%%
      for n=1:N
            complexity_counter(1) = complexity_counter(1) + 1;
          if Irow(n)==i || Irow(n)==j || n==i || n==j
                complexity_counter(2) = complexity_counter(2) + 1;
              [f(n),Irow(n)]=max(F(:,n));
          else
                complexity_counter(3) = complexity_counter(3) + 1;
              if f(n)<F(i,n)
                  f(n)=F(i,n);
                  Irow(n)=i;
              end
              if f(n)<F(j,n)
                  f(n)=F(j,n);
                  Irow(n)=j;
              end
          end
      end


    %  MDL(iterations)=M*N/2+M/2*sum(log(diag(R)))+N*M/2*log(2*pi)+iterations/2*log(N*M);

      if flag_design==1 % && mod(iterations,ceil(NumRotations/4))==0
            % S=A'*S*A 
            ci = S(:,i)*cos(theta)+S(:,j)*sin(theta);   % R*A
            cj = -S(:,i)*sin(theta)+S(:,j)*cos(theta);
            S(:,i) = ci;
            S(:,j) = cj;

            ri = S(i,:)*cos(theta) + S(j,:)*sin(theta); % A'*R*A
            rj = -S(i,:)*sin(theta) + S(j,:)*cos(theta); 
            S(i,:) = ri;  
            S(j,:) = rj;

            % average log-likelihood of a single data vector
            likelihood(iterations) = -1/2*sum(diag(S)./diag(R))-1/2*sum(log(diag(R)))-N/2*log(2*pi);      
      end


      %%%%% sparse representation of Givens rotations %%%%%%
      SMTArray(iterations,1)=i;
      SMTArray(iterations,2)=j;
      SMTArray(iterations,3)=theta;
      lambda=diag(R);
    end


function F = Criteria(R)
	lambda = diag(R);
	F = (R-diag(lambda)).^2 ./ (lambda*lambda');
    

function [a,b,theta] = rotation(rii,rjj,rij)
	tmp1 = 0.5*(rii + rjj);
	tmp2 = 0.5*sqrt( (rii-rjj)^2 + 4*rij^2 );

	a = tmp1 + tmp2;
	b = tmp1 - tmp2;
    theta = (atan2(2*rij,rii-rjj))/2;

